# -*- coding: utf-8 -*-
import os
import yaml
from pathlib import Path

__all__ = ['CoreConfig']


def load_config(config_file_dir: str = 'CoreConfig.yaml'):
    config_yaml = list(Path().rglob(config_file_dir))

    if len(config_yaml) == 1:
        print(f"[Config] File Exist → {config_yaml[0].resolve()}")
        with open(config_yaml[0], encoding="utf-8") as f:
            data = yaml.load(f, Loader=yaml.SafeLoader)
    else:
        print("[Config] File Not Found")
        data = {
            "connections": {
                "base": {
                    'engine': 'tortoise.backends.mysql',
                    "credentials": {
                        'host': os.getenv('BASE_HOST', '127.0.0.1'),
                        'user': os.getenv('BASE_USER', 'root'),
                        'password': os.getenv('BASE_PASSWORD', '123456'),
                        'port': int(os.getenv('BASE_PORT', 3306)),
                        'database': os.getenv('BASE_DB', 'base'),
                    }
                },
            },
            "apps": {
                "base": {"models": ["models.base"], "default_connection": "base"},
                # "db2": {"models": ["models.db2"], "default_connection": "db2"},
                # "db3": {"models": ["models.db3"], "default_connection": "db3"}
            },
            'use_tz': False,
            'timezone': 'Asia/Shanghai'
        }
    return AuConfig(**data)


class AuConfig:
    waveform = {"type": ".wav",
                "default_rate": 44100,
                "common_rate": [192000, 96000, 88200, 64000, 48000, 44100, 32000, 22050, 16000, 11025, 8000, 6000],
                "resample_rate": 16000,
                "channels": 'Mono',
                "format": '16bit',
                }
    device = {'pool_max': 200, 'cpu_cores': 8}

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


# a = AuConfig()
CoreConfig = load_config('CoreConfig.yaml')
