# -*- coding: utf-8 -*-

import os
import time
import pandas as pd
from collections import Counter
from matplotlib import pyplot as plt

__all__ = [
    'data2xlsx',
    'lowercase',
    'upercase',
    'find_duplicate',
    'history_plot',
    'fRed',
    'fGre',
    'red',
    'gre',
    'blu',
    'yel',
    'pur',
    'cya',
    'dkred',
    'dkgre',
    'dkblu',
    'dkyel',
    'dkpur',
    'dkcya',
    'h1',
    'info',
    'warn',
    'error',
    'timer',
    'timeEnd',
    'wait',
    'notify',
    'tsp',
]


def data2xlsx(data, xlsx_dir, sheet_name):
    if not pd.__version__ >= '1.0.3':
        warn("Need pandas version >= 1.0.3")
        return
    if isinstance(data, dict):
        try:
            data_frame = pd.DataFrame.from_dict(data, orient='columns')
        except Exception as e:
            print(e)
            return f'[data2xlsx] data error: {e}'
    elif isinstance(data, pd.DataFrame):
        data_frame = data
    else:
        return '[data2xlsx] check the data type'
    if not os.path.isdir(os.path.dirname(xlsx_dir)):
        os.makedirs(os.path.dirname(xlsx_dir))
    # write excel (pandas==1.0.3)
    if os.path.isfile(xlsx_dir):
        with pd.ExcelWriter(xlsx_dir, engine='openpyxl', mode='a') as writer:
            data_frame.to_excel(writer, sheet_name=sheet_name, index=False)
    else:
        data_frame.to_excel(xlsx_dir, sheet_name=sheet_name, index=False)


def lowercase(list):
    return [_.lower() for _ in list if isinstance(_, str)]


def upercase(list):
    return [_.upper() for _ in list if isinstance(_, str)]


def find_duplicate(file_list):
    b = dict(Counter(file_list))
    print([key for key, value in b.items() if value > 1])  # 只展示重複元素
    print({key: value for key, value in b.items() if value > 1})  # 展現重複元素和重複次數
    return b


def history_plot(history, metrics='loss', save_fig=None):
    # plt.figure()
    for_legend = []
    color = ['r', 'g', 'b', 'c', 'm', 'y', 'k', 'b--', 'r:', 'c:', 'm:', 'k:', 'g--', 'y--']
    his_list = list(history.history.keys())
    for index, h_list in enumerate(his_list):
        if h_list.find(metrics) > -1:
            valid = history.history.get(h_list)
            plt.plot(valid, color[index], label='h_list')
            for_legend.append(h_list)
    plt.legend(for_legend, loc='upper left', fontsize='small')
    plt.title(metrics)
    plt.xlabel('Epoch')
    plt.ylabel(metrics)
    plt.grid(True)
    if save_fig:
        if not os.path.isdir(os.path.basename(save_fig)):
            os.makedirs(os.path.basename(save_fig))
        if save_fig.endswith('.png'):
            plt.savefig("{}_{}".format(metrics, save_fig), dpi=150)
        else:
            plt.savefig("{}_{}.png".format(metrics, save_fig), dpi=150)
    # plt.show()
    plt.close()


class Colors:
    """
    顯示方式：
        預設值：0
        高亮：1
        非粗體：22
        下劃線：4
        非下劃線：24
        閃爍：5
        非閃爍：25
        反顯：7
        非反顯：27

    前景色：字型顯示的顏色
        黑色：30
        紅色：31
        綠色：32
        黃色：33
        藍色：34
        洋紅：35
        青色：36
        白色：37

    背景色：字型的背景顏色
        黑色：40
        紅色：41
        綠色：42
        黃色：43
        藍色：44
        洋紅：45
        青色：46
        白色：47
    """
    END = "\033[0m"

    BRIGHT = "\033[1m"
    DIM = "\033[2m"
    UNDERSCORE = "\033[4m"
    BLINK = "\033[5m"

    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    BACK_RED = "\033[41m"
    BACK_GREEN = "\033[42m"
    BACK_YELLOW = "\033[43m"
    BACK_BLUE = "\033[44m"
    BACK_MAGENTA = "\033[45m"
    BACK_CYAN = "\033[46m"
    BACK_WHITE = "\033[47m"

    BRIGHT_WHHITE_BACK_BLUE = "\033[1;44;37m"
    BRIGHT_GREEN_BACK_BLUE = "\033[1;42;37m"
    BRIGHT_RED_BACK_BLUE = "\033[1;41;37m"


timers = {}


def fmt(iterable):
    return " ".join(str(i) for i in iterable)


def fRed(*args):
    print(Colors.BLINK + Colors.WHITE + Colors.BACK_RED, fmt(args), Colors.END)


def fGre(*args):
    print(Colors.BLINK + Colors.WHITE + Colors.BACK_GREEN, fmt(args), Colors.END)


def red(*args):
    print(Colors.RED, fmt(args), Colors.END)


def gre(*args):
    print(Colors.GREEN, fmt(args), Colors.END)


def blu(*args):
    print(Colors.BLUE, fmt(args), Colors.END)


def yel(*args):
    print(Colors.YELLOW, fmt(args), Colors.END)


def pur(*args):
    print(Colors.MAGENTA, fmt(args), Colors.END)


def cya(*args):
    print(Colors.CYAN, fmt(args), Colors.END)


def dkred(*args):
    print(Colors.BACK_RED, fmt(args), Colors.END)


def dkgre(*args):
    print(Colors.BACK_GREEN, fmt(args), Colors.END)


def dkblu(*args):
    print(Colors.BACK_BLUE, fmt(args), Colors.END)


def dkyel(*args):
    print(Colors.BACK_YELLOW, fmt(args), Colors.END)


def dkpur(*args):
    print(Colors.BACK_MAGENTA, fmt(args), Colors.END)


def dkcya(*args):
    print(Colors.BACK_CYAN, fmt(args), Colors.END)


def h1(*args):
    print(Colors.BRIGHT, fmt(args), Colors.END)


def info(*args):
    print(Colors.DIM + "\t", fmt(args), Colors.END)


def warn(*args):
    print(Colors.BACK_CYAN + "WARN:" + Colors.END + Colors.CYAN, fmt(args), Colors.END)


def error(*args):
    print(Colors.BACK_RED + Colors.BLINK + "ERROR:" + Colors.END + Colors.RED, fmt(args), Colors.END)


def timer(key):
    timers[key] = time.time()


def timeEnd(key):
    if key in timers:
        t = time.time() - timers[key]
        print(key + "\t" + str(t) + Colors.DIM + " s " + Colors.END)
        del timers[key]


def wait(*args):
    return str.lower(input(Colors.BLUE + fmt(args) + Colors.END))


def notify(*args):
    # Play bell
    print('\a')
    # Attempt to send a notification (will fail, but not crash, if not on macOS)
    os.system("""
          osascript -e 'display notification "{}" with title "{}"'
          """.format(args[0], fmt(args[1:])))


def tsp(w: str = None):
    if w:
        print(f"\033[1;42;37m[GD] ➡GPIO {w} | Out \033[0m")
        gre(f"[GD] ➡GPIO {w} | Out ")
        print(f"<font color = 'orange'>➡GPIO {w} | Out </font>")
    else:
        print("\033[1;42;37m[GD] ➡GPIO 123 | Out \033[0m")
        gre("[GD] ➡GPIO 456 | Out ")
        print("<font color = 'orange'>➡GPIO {pin} | Out </font>")
